@@ -10,7 +10,6 @@ import (
 	"k8s.io/apimachinery/pkg/api/errors"
 	k8serror "k8s.io/apimachinery/pkg/api/errors"
 	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
-	"k8s.io/apimachinery/pkg/types"
 	k8sclient "sigs.k8s.io/controller-runtime/pkg/client"
 )
 
@@ -38,9 +37,6 @@ const (
 	// Resources included in the stage backup.
 	// Referenced by the Backup.LabelSelector. The value is the Task.UID().
 	IncludedInStageBackupLabel = "migration-included-stage-backup"
-	// Application pods requiring restic/stage backups.
-	// Used to create stage pods. The value is the Task.UID()
-	ApplicationPodLabel = "migration-application-pod"
 	// Designated as an `initial` Backup.
 	// The value is the Task.UID().
 	InitialBackupLabel = "migration-initial-backup"
@@ -107,13 +103,8 @@ func (t *Task) annotateStageResources() error {
 	return nil
 }
 
-// Add annotations and labels to PVCs.
-// The PvActionAnnotation annotation is added to PVCs as needed by the velero plugin.
-// The PvStorageClassAnnotation annotation is added to PVC as needed by the velero plugin.
-// The PvAccessModeAnnotation annotation is added to PVC as needed by the velero plugin.
-// The IncludedInStageBackupLabel label is added to PVCs and is referenced
-// by the velero.Backup label selector.
-func (t *Task) annotatePVCs(client k8sclient.Client, pod corev1.Pod) ([]string, []string, error) {
+// Gets a list of restic volumes and restic verify volumes for a pod
+func (t *Task) getResticVolumes(client k8sclient.Client, pod corev1.Pod) ([]string, []string, error) {
 	volumes := []string{}
 	verifyVolumes := []string{}
 	pvs := t.getPVs()
@@ -125,22 +116,16 @@ func (t *Task) annotatePVCs(client k8sclient.Client, pod corev1.Pod) ([]string,
 		pvc := corev1.PersistentVolumeClaim{}
 		err := client.Get(
 			context.TODO(),
-			types.NamespacedName{
+			k8sclient.ObjectKey{
 				Namespace: pod.Namespace,
 				Name:      claim.ClaimName,
 			},
 			&pvc)
-		// PV action (move|copy) annotation needed by the velero plugin.
-		if pvc.Annotations == nil {
-			pvc.Annotations = make(map[string]string)
+		if err != nil {
+			log.Trace(err)
+			return nil, nil, err
 		}
 		pvAction := findPVAction(pvs, pvc.Spec.VolumeName)
-		pvc.Annotations[PvActionAnnotation] = pvAction
-		// Add label used by stage backup label selector.
-		if pvc.Labels == nil {
-			pvc.Labels = make(map[string]string)
-		}
-		pvc.Labels[IncludedInStageBackupLabel] = t.UID()
 		if pvAction == migapi.PvCopyAction {
 			// Add to Restic volume list if copyMethod is "filesystem"
 			if findPVCopyMethod(pvs, pvc.Spec.VolumeName) == migapi.PvFilesystemCopyMethod {
@@ -149,29 +134,8 @@ func (t *Task) annotatePVCs(client k8sclient.Client, pod corev1.Pod) ([]string,
 					verifyVolumes = append(verifyVolumes, pv.Name)
 				}
 			}
-			// PV storageClass annotation needed by the velero plugin.
-			storageClass := findPVStorageClass(pvs, pvc.Spec.VolumeName)
-			pvc.Annotations[PvStorageClassAnnotation] = storageClass
-			// PV accessMode annotation needed by the velero plugin, if present on the PV.
-			accessMode := findPVAccessMode(pvs, pvc.Spec.VolumeName)
-			if accessMode != "" {
-				pvc.Annotations[PvAccessModeAnnotation] = string(accessMode)
-			}
 		}
-		// Update
-		err = client.Update(context.TODO(), &pvc)
-		if err != nil {
-			log.Trace(err)
-			return nil, nil, err
-		}
-		log.Info(
-			"PVC annotations/labels added.",
-			"ns",
-			pvc.Namespace,
-			"name",
-			pvc.Name)
 	}
-
 	return volumes, verifyVolumes, nil
 }
 
@@ -181,7 +145,7 @@ func (t *Task) labelNamespaces(client k8sclient.Client) error {
 		namespace := corev1.Namespace{}
 		err := client.Get(
 			context.TODO(),
-			types.NamespacedName{
+			k8sclient.ObjectKey{
 				Name: ns,
 			},
 			&namespace)
@@ -214,105 +178,85 @@ func (t *Task) labelNamespaces(client k8sclient.Client) error {
 // Returns a set of referenced service accounts.
 func (t *Task) annotatePods(client k8sclient.Client) (ServiceAccounts, error) {
 	serviceAccounts := ServiceAccounts{}
-	for _, ns := range t.sourceNamespaces() {
-		list := corev1.PodList{}
-		options := k8sclient.InNamespace(ns)
-		err := client.List(context.TODO(), options, &list)
+	list := corev1.PodList{}
+	// include stage pods only
+	options := k8sclient.MatchingLabels(t.Owner.GetCorrelationLabels())
+	err := client.List(context.TODO(), options, &list)
+	if err != nil {
+		log.Trace(err)
+		return nil, err
+	}
+	for _, pod := range list.Items {
+		// Annotate PVCs.
+		volumes, verifyVolumes, err := t.getResticVolumes(client, pod)
 		if err != nil {
 			log.Trace(err)
 			return nil, err
 		}
-		for _, pod := range list.Items {
-			if pod.Labels == nil {
-				pod.Labels = make(map[string]string)
-			}
-			// Skip stage pods.
-			cLabel, _ := t.Owner.GetCorrelationLabel()
-			if _, found := pod.Labels[cLabel]; found {
-				continue
-			}
-			// Skip stateless pods.
-			if len(pod.Spec.Volumes) == 0 {
-				continue
-			}
-			// Annotate PVCs.
-			volumes, verifyVolumes, err := t.annotatePVCs(client, pod)
-			if err != nil {
-				log.Trace(err)
-				return nil, err
-			}
-			if len(volumes) == 0 {
-				continue
-			}
-			// Restic annotation used to specify volumes.
-			if pod.Annotations == nil {
-				pod.Annotations = make(map[string]string)
-			}
-			pod.Annotations[ResticPvBackupAnnotation] = strings.Join(volumes, ",")
-			pod.Annotations[ResticPvVerifyAnnotation] = strings.Join(verifyVolumes, ",")
-			// Add label used by stage backup label selector.
-			if pod.Labels == nil {
-				pod.Labels = make(map[string]string)
-			}
-			pod.Labels[ApplicationPodLabel] = t.UID()
-			// Update
-			err = client.Update(context.TODO(), &pod)
-			if err != nil {
-				log.Trace(err)
-				return nil, err
-			}
-			log.Info(
-				"Pod annotations/labels added.",
-				"ns",
-				pod.Namespace,
-				"name",
-				pod.Name)
-			sa := pod.Spec.ServiceAccountName
-			names, found := serviceAccounts[pod.Namespace]
-			if !found {
-				serviceAccounts[pod.Namespace] = map[string]bool{sa: true}
-			} else {
-				names[sa] = true
-			}
+		// Restic annotation used to specify volumes.
+		if pod.Annotations == nil {
+			pod.Annotations = make(map[string]string)
+		}
+		pod.Annotations[ResticPvBackupAnnotation] = strings.Join(volumes, ",")
+		pod.Annotations[ResticPvVerifyAnnotation] = strings.Join(verifyVolumes, ",")
+		// Update
+		err = client.Update(context.TODO(), &pod)
+		if err != nil {
+			log.Trace(err)
+			return nil, err
+		}
+		log.Info(
+			"Pod annotations/labels added.",
+			"ns",
+			pod.Namespace,
+			"name",
+			pod.Name)
+		sa := pod.Spec.ServiceAccountName
+		names, found := serviceAccounts[pod.Namespace]
+		if !found {
+			serviceAccounts[pod.Namespace] = map[string]bool{sa: true}
+		} else {
+			names[sa] = true
 		}
 	}
 
 	return serviceAccounts, nil
 }
 
-// Add annotations and labels to PVs.
-// The PvActionAnnotation annotation is added to PVs as needed by the velero plugin.
-// The PvStorageClassAnnotation annotation is added to PVs as needed by the velero plugin.
-// The IncludedInStageBackupLabel label is added to PVs and is referenced
+// Add annotations and labels to PVs and PVCs.
+// The PvActionAnnotation annotation is added to PVs and PVCs as needed by the velero plugin.
+// The PvStorageClassAnnotation annotation is added to PVs and PVCs as needed by the velero plugin.
+// The IncludedInStageBackupLabel label is added to PVs and PVCs and is referenced
 // by the velero.Backup label selector.
+// The PvAccessModeAnnotation annotation is added to PVC as needed by the velero plugin.
 func (t *Task) annotatePVs(client k8sclient.Client) error {
 	pvs := t.getPVs()
 	for _, pv := range pvs.List {
-		resource := corev1.PersistentVolume{}
+		pvResource := corev1.PersistentVolume{}
 		err := client.Get(
 			context.TODO(),
-			types.NamespacedName{
+			k8sclient.ObjectKey{
 				Name: pv.Name,
 			},
-			&resource)
+			&pvResource)
 		if err != nil {
 			log.Trace(err)
 			return err
 		}
-		if resource.Annotations == nil {
-			resource.Annotations = make(map[string]string)
+		if pvResource.Annotations == nil {
+			pvResource.Annotations = make(map[string]string)
 		}
 		// PV action (move|copy) needed by the velero plugin.
-		resource.Annotations[PvActionAnnotation] = pv.Selection.Action
+		pvResource.Annotations[PvActionAnnotation] = pv.Selection.Action
 		// PV storageClass annotation needed by the velero plugin.
-		resource.Annotations[PvStorageClassAnnotation] = pv.Selection.StorageClass
+		pvResource.Annotations[PvStorageClassAnnotation] = pv.Selection.StorageClass
 		// Add label used by stage backup label selector.
-		if resource.Labels == nil {
-			resource.Labels = make(map[string]string)
+		if pvResource.Labels == nil {
+			pvResource.Labels = make(map[string]string)
 		}
-		resource.Labels[IncludedInStageBackupLabel] = t.UID()
+		pvResource.Labels[IncludedInStageBackupLabel] = t.UID()
 		// Update
-		err = client.Update(context.TODO(), &resource)
+		err = client.Update(context.TODO(), &pvResource)
 		if err != nil {
 			log.Trace(err)
 			return err
@@ -321,6 +265,49 @@ func (t *Task) annotatePVs(client k8sclient.Client) error {
 			"PV annotations/labels added.",
 			"name",
 			pv.Name)
+
+		pvcResource := corev1.PersistentVolumeClaim{}
+		err = client.Get(
+			context.TODO(),
+			k8sclient.ObjectKey{
+				Namespace: pv.PVC.Namespace,
+				Name:      pv.PVC.Name,
+			},
+			&pvcResource)
+		if err != nil {
+			log.Trace(err)
+			return err
+		}
+		if pvcResource.Annotations == nil {
+			pvcResource.Annotations = make(map[string]string)
+		}
+		// PV action (move|copy) needed by the velero plugin.
+		pvcResource.Annotations[PvActionAnnotation] = pv.Selection.Action
+		// Add label used by stage backup label selector.
+		if pvcResource.Labels == nil {
+			pvcResource.Labels = make(map[string]string)
+		}
+		pvcResource.Labels[IncludedInStageBackupLabel] = t.UID()
+		if pv.Selection.Action == migapi.PvCopyAction {
+			// PV storageClass annotation needed by the velero plugin.
+			pvcResource.Annotations[PvStorageClassAnnotation] = pv.Selection.StorageClass
+			// PV accessMode annotation needed by the velero plugin, if present on the PV.
+			if pv.Selection.AccessMode != "" {
+				pvcResource.Annotations[PvAccessModeAnnotation] = string(pv.Selection.AccessMode)
+			}
+		}
+		// Update
+		err = client.Update(context.TODO(), &pvcResource)
+		if err != nil {
+			log.Trace(err)
+			return err
+		}
+		log.Info(
+			"PVC annotations/labels added.",
+			"ns",
+			pv.PVC.Namespace,
+			"name",
+			pv.PVC.Name)
 	}
 
 	return nil
@@ -434,10 +421,6 @@ func (t *Task) deletePodAnnotations(client k8sclient.Client, namespaceList []str
 					delete(pod.Labels, IncludedInStageBackupLabel)
 					needsUpdate = true
 				}
-				if _, found := pod.Labels[ApplicationPodLabel]; found {
-					delete(pod.Labels, ApplicationPodLabel)
-					needsUpdate = true
-				}
 			}
 			if !needsUpdate {
 				continue
@@ -465,7 +448,7 @@ func (t *Task) deleteNamespaceLabels(client k8sclient.Client, namespaceList []st
 		namespace := corev1.Namespace{}
 		err := client.Get(
 			context.TODO(),
-			types.NamespacedName{
+			k8sclient.ObjectKey{
 				Name: ns,
 			},
 			&namespace)
@@ -540,6 +523,9 @@ func (t *Task) deletePVCAnnotations(client k8sclient.Client, namespaceList []str
 			return err
 		}
 		for _, pvc := range pvcList.Items {
+			if pvc.Spec.VolumeName == "" {
+				continue
+			}
 			needsUpdate := false
 			if pvc.Annotations != nil {
 				if _, found := pvc.Annotations[PvActionAnnotation]; found {
